@@ -1,4 +1,4 @@
-# Copyright 2021-2022 NVIDIA Corporation.  All rights reserved.
+# Copyright 2021-2024 NVIDIA Corporation.  All rights reserved.
 #
 # Please refer to the NVIDIA end user license agreement (EULA) associated
 # with this source code for terms and conditions that govern your use of
@@ -29,6 +29,9 @@ def supportsManagedMemory():
     err, isSupported = cudart.cudaDeviceGetAttribute(cudart.cudaDeviceAttr.cudaDevAttrManagedMemory, 0)
     return err == cudart.cudaError_t.cudaSuccess and isSupported
 
+def supportsCudaAPI(name):
+    return name in dir(cuda)
+
 def callableBinary(name):
     return shutil.which(name) != None
 
@@ -499,7 +502,7 @@ def test_cuda_graphMem_attr():
     err, freeNode = cuda.cuGraphAddMemFreeNode(graph, [allocNode], 1, params.dptr)
     assert(err == cuda.CUresult.CUDA_SUCCESS)
 
-    err, graphExec, _ = cuda.cuGraphInstantiate(graph, b'', 0)
+    err, graphExec = cuda.cuGraphInstantiate(graph, 0)
     assert(err == cuda.CUresult.CUDA_SUCCESS)
 
     err, = cuda.cuGraphLaunch(graphExec, stream)
@@ -526,6 +529,45 @@ def test_cuda_graphMem_attr():
     err, = cuda.cuCtxDestroy(ctx)
     assert(err == cuda.CUresult.CUDA_SUCCESS)
 
+@pytest.mark.skipif(driverVersionLessThan(12010)
+                    or not supportsCudaAPI('cuCoredumpSetAttributeGlobal')
+                    or not supportsCudaAPI('cuCoredumpGetAttributeGlobal'), reason='Coredump API not present')
+def test_cuda_coredump_attr():
+    err, = cuda.cuInit(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, device = cuda.cuDeviceGet(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, ctx = cuda.cuCtxCreate(0, device)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    attr_list = [None] * 6
+
+    err, = cuda.cuCoredumpSetAttributeGlobal(cuda.CUcoredumpSettings.CU_COREDUMP_TRIGGER_HOST, False)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuCoredumpSetAttributeGlobal(cuda.CUcoredumpSettings.CU_COREDUMP_FILE, b'corefile')
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuCoredumpSetAttributeGlobal(cuda.CUcoredumpSettings.CU_COREDUMP_PIPE, b'corepipe')
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuCoredumpSetAttributeGlobal(cuda.CUcoredumpSettings.CU_COREDUMP_LIGHTWEIGHT, True)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    for idx, attr in enumerate([cuda.CUcoredumpSettings.CU_COREDUMP_TRIGGER_HOST,
+                                cuda.CUcoredumpSettings.CU_COREDUMP_FILE,
+                                cuda.CUcoredumpSettings.CU_COREDUMP_PIPE,
+                                cuda.CUcoredumpSettings.CU_COREDUMP_LIGHTWEIGHT,
+                                ]):
+        err, attr_tmp = cuda.cuCoredumpGetAttributeGlobal(attr)
+        assert(err == cuda.CUresult.CUDA_SUCCESS)
+        attr_list[idx] = attr_tmp
+
+    assert(attr_list[0] == False)
+    assert(attr_list[1] == b'corefile')
+    assert(attr_list[2] == b'corepipe')
+    assert(attr_list[3] == True)
+
+    err, = cuda.cuCtxDestroy(ctx)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
 def test_get_error_name_and_string():
     err, = cuda.cuInit(0)
     assert(err == cuda.CUresult.CUDA_SUCCESS)
@@ -567,11 +609,14 @@ def test_device_get_name():
     delimiter = b'\r\n' if platform.system() == "Windows" else b'\n'
     expect = p.stdout.split(delimiter)
     size = 64
-    _, got = cuda.cuDeviceGetName(size, cuda.CUdevice(0))
-    # Returned value is bytes, and we expect it to be of requested size
-    # assert len(got) == size
+    _, got = cuda.cuDeviceGetName(size, device)
     got = got.split(b'\x00')[0]
-    assert got in expect
+    if any(b'Unable to determine the device handle for' in result for result in expect):
+        # Undeterministic devices get waived
+        pass
+    else:
+        assert any(got in result for result in expect)
+
 
     err, = cuda.cuCtxDestroy(ctx)
     assert(err == cuda.CUresult.CUDA_SUCCESS)
@@ -581,19 +626,6 @@ def test_device_get_name():
 def test_stream_capture():
     pass
 
-def test_c_func_callback():
-    err, = cuda.cuInit(0)
-    assert(err == cuda.CUresult.CUDA_SUCCESS)
-    err, device = cuda.cuDeviceGet(0)
-    assert(err == cuda.CUresult.CUDA_SUCCESS)
-    err, ctx = cuda.cuCtxCreate(0, device)
-    assert(err == cuda.CUresult.CUDA_SUCCESS)
-
-    # TODO
-
-    err, = cuda.cuCtxDestroy(ctx)
-    assert(err == cuda.CUresult.CUDA_SUCCESS)
-
 def test_profiler():
     err, = cuda.cuInit(0)
     assert(err == cuda.CUresult.CUDA_SUCCESS)
@@ -624,7 +656,7 @@ def test_eglFrame():
     assert(int(val.frame.pArray[0]) == 4)
     assert(int(val.frame.pArray[1]) == 2)
     assert(int(val.frame.pArray[2]) == 3)
-    val.frame.pPitch
+    val.frame.pPitch = [4, 2, 3]
     # [4, 2, 3]
     assert(int(val.frame.pPitch[0]) == 4)
     assert(int(val.frame.pPitch[1]) == 2)
@@ -633,3 +665,211 @@ def test_eglFrame():
     assert(int(val.frame.pPitch[0]) == 1)
     assert(int(val.frame.pPitch[1]) == 2)
     assert(int(val.frame.pPitch[2]) == 3)
+
+def test_char_range():
+    val = cuda.CUipcMemHandle_st()
+    for x in range(-128, 0):
+        val.reserved = [x] * 64
+        assert(val.reserved[0] == 256 + x)
+    for x in range(0, 256):
+        val.reserved = [x] * 64
+        assert(val.reserved[0] == x)
+
+def test_anon_assign():
+    val1 = cuda.CUexecAffinityParam_st()
+    val2 = cuda.CUexecAffinityParam_st()
+
+    assert(val1.param.smCount.val == 0)
+    val1.param.smCount.val = 5
+    assert(val1.param.smCount.val == 5)
+    val2.param.smCount.val = 11
+    assert(val2.param.smCount.val == 11)
+
+    val1.param = val2.param
+    assert(val1.param.smCount.val == 11)
+
+def test_union_assign():
+    val = cuda.CUlaunchAttributeValue()
+    val.clusterDim.x, val.clusterDim.y, val.clusterDim.z = 9,9,9
+    attr = cuda.CUlaunchAttribute()
+    attr.value = val
+
+    assert(val.clusterDim.x == 9)
+    assert(val.clusterDim.y == 9)
+    assert(val.clusterDim.z == 9)
+
+def test_invalid_repr_attribute():
+    val = cuda.CUlaunchAttributeValue()
+    string = str(val)
+
+@pytest.mark.skipif(driverVersionLessThan(12020)
+                    or not supportsCudaAPI('cuGraphAddNode')
+                    or not supportsCudaAPI('cuGraphNodeSetParams')
+                    or not supportsCudaAPI('cuGraphExecNodeSetParams'), reason='Polymorphic graph APIs required')
+def test_graph_poly():
+    err, = cuda.cuInit(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, device = cuda.cuDeviceGet(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, ctx = cuda.cuCtxCreate(0, device)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, stream = cuda.cuStreamCreate(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    # cuGraphAddNode
+
+    # Create 2 buffers
+    size = int(1024 * np.uint8().itemsize)
+    buffers = []
+    for _ in range(2):
+        err, dptr = cuda.cuMemAlloc(size)
+        assert(err == cuda.CUresult.CUDA_SUCCESS)
+        buffers += [(np.full(size, 2).astype(np.uint8), dptr)]
+
+    # Update dev buffers
+    for host, device in buffers:
+        err, = cuda.cuMemcpyHtoD(device, host, size)
+        assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    # Create graph
+    nodes = []
+    err, graph = cuda.cuGraphCreate(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    # Memset
+    host, device = buffers[0]
+    memsetParams = cuda.CUgraphNodeParams()
+    memsetParams.type = cuda.CUgraphNodeType.CU_GRAPH_NODE_TYPE_MEMSET
+    memsetParams.memset.elementSize = np.uint8().itemsize
+    memsetParams.memset.width = size
+    memsetParams.memset.height = 1
+    memsetParams.memset.dst = device
+    memsetParams.memset.value = 1
+    err, node = cuda.cuGraphAddNode(graph, None, 0, memsetParams)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    nodes += [node]
+
+    # Memcpy
+    host, device = buffers[1]
+    memcpyParams = cuda.CUgraphNodeParams()
+    memcpyParams.type = cuda.CUgraphNodeType.CU_GRAPH_NODE_TYPE_MEMCPY
+    memcpyParams.memcpy.copyParams.srcMemoryType = cuda.CUmemorytype.CU_MEMORYTYPE_DEVICE
+    memcpyParams.memcpy.copyParams.srcDevice = device
+    memcpyParams.memcpy.copyParams.dstMemoryType = cuda.CUmemorytype.CU_MEMORYTYPE_HOST
+    memcpyParams.memcpy.copyParams.dstHost = host
+    memcpyParams.memcpy.copyParams.WidthInBytes = size
+    memcpyParams.memcpy.copyParams.Height = 1
+    memcpyParams.memcpy.copyParams.Depth = 1
+    err, node = cuda.cuGraphAddNode(graph, None, 0, memcpyParams)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    nodes += [node]
+
+    # Instantiate, execute, validate
+    err, graphExec = cuda.cuGraphInstantiate(graph, 0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuGraphLaunch(graphExec, stream)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuStreamSynchronize(stream)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    # Validate
+    for host, device in buffers:
+        err, = cuda.cuMemcpyDtoH(host, device, size)
+        assert(err == cuda.CUresult.CUDA_SUCCESS)
+    assert(np.array_equal(buffers[0][0], np.full(size, 1).astype(np.uint8)))
+    assert(np.array_equal(buffers[1][0], np.full(size, 2).astype(np.uint8)))
+
+    # cuGraphNodeSetParams
+    host, device = buffers[1]
+    err, memcpyParamsCopy = cuda.cuGraphMemcpyNodeGetParams(nodes[1])
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    assert(int(memcpyParamsCopy.srcDevice) == int(device))
+    host, device = buffers[0]
+    memcpyParams.memcpy.copyParams.srcDevice = device
+    err, = cuda.cuGraphNodeSetParams(nodes[1], memcpyParams)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, memcpyParamsCopy = cuda.cuGraphMemcpyNodeGetParams(nodes[1])
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    assert(int(memcpyParamsCopy.srcDevice) == int(device))
+
+    # cuGraphExecNodeSetParams
+    memsetParams.memset.value = 11
+    err, = cuda.cuGraphExecNodeSetParams(graphExec, nodes[0], memsetParams)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuGraphLaunch(graphExec, stream)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuStreamSynchronize(stream)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuMemcpyDtoH(buffers[0][0], buffers[0][1], size)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    assert(np.array_equal(buffers[0][0], np.full(size, 11).astype(np.uint8)))
+
+    # Cleanup
+    err, = cuda.cuMemFree(buffers[0][1])
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuMemFree(buffers[1][1])
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuGraphExecDestroy(graphExec)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuGraphDestroy(graph)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuStreamDestroy(stream)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, = cuda.cuCtxDestroy(ctx)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+@pytest.mark.skipif(driverVersionLessThan(12040)
+                    or not supportsCudaAPI('cuDeviceGetDevResource'), reason='Polymorphic graph APIs required')
+def test_cuDeviceGetDevResource():
+    err, = cuda.cuInit(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, device = cuda.cuDeviceGet(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, resource_in = cuda.cuDeviceGetDevResource(device, cuda.CUdevResourceType.CU_DEV_RESOURCE_TYPE_SM)
+    err, ctx = cuda.cuCtxCreate(0, device)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    err, res, count, rem = cuda.cuDevSmResourceSplitByCount(0, resource_in, 0, 2)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    assert(count != 0)
+    assert(len(res) == 0)
+    err, res, count_same, rem = cuda.cuDevSmResourceSplitByCount(count, resource_in, 0, 2)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    assert(count == count_same)
+    assert(len(res) == count)
+    err, res, count, rem = cuda.cuDevSmResourceSplitByCount(3, resource_in, 0, 2)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    assert(len(res) == 3)
+
+    err, = cuda.cuCtxDestroy(ctx)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+@pytest.mark.skipif(driverVersionLessThan(12030)
+                    or not supportsCudaAPI('cuGraphConditionalHandleCreate'), reason='Conditional graph APIs required')
+def test_conditional():
+    err, = cuda.cuInit(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, device = cuda.cuDeviceGet(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, ctx = cuda.cuCtxCreate(0, device)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    err, graph = cuda.cuGraphCreate(0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+    err, handle = cuda.cuGraphConditionalHandleCreate(graph, ctx, 0, 0)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    params = cuda.CUgraphNodeParams()
+    params.type = cuda.CUgraphNodeType.CU_GRAPH_NODE_TYPE_CONDITIONAL
+    params.conditional.handle = handle
+    params.conditional.type = cuda.CUgraphConditionalNodeType.CU_GRAPH_COND_TYPE_IF
+    params.conditional.size = 1
+    params.conditional.ctx = ctx
+
+    assert(len(params.conditional.phGraph_out) == 1)
+    assert(int(params.conditional.phGraph_out[0]) == 0)
+    err, node = cuda.cuGraphAddNode(graph, None, 0, params)
+    assert(err == cuda.CUresult.CUDA_SUCCESS)
+
+    assert(len(params.conditional.phGraph_out) == 1)
+    assert(int(params.conditional.phGraph_out[0]) != 0)
